#include "Cluster.h"

Cluster::Cluster(): pivot(0), radius(numeric_limits<SimType>::min()) {
  index=new ContCluster;
  ftable=new FTable();
}

Cluster::Cluster(const Cluster& c): pivot(c.pivot), radius(c.radius) {
  index=new ContCluster;
  *index=*c.index;
  ftable=new FTable;
  *ftable=*c.ftable;
}

Cluster::~Cluster() {
  delete index;
  delete ftable;
}

Cluster& Cluster::operator=(const Cluster& c) {
  if (this==&c)
    return *this;
  *index=*c.index;
  *ftable=*c.ftable;
  pivot=c.pivot;
  radius=c.radius;
  return *this;
}

void Cluster::set_pivot(const unsigned& p) {
  pivot=p;
}

void Cluster::set_radius(const SimType r) {
  radius=r;
}

const size_t Cluster::size() const {
  return this->index->size();
}

const unsigned Cluster::get_pivot() const {
  return pivot;
}

const SimType Cluster::get_radius() const {
  return radius;
}

const FTable& Cluster::get_ftable() const {
  return *ftable;
}

ContCluster::const_iterator Cluster::begin() const {
  return index->begin();
}

ContCluster::const_iterator Cluster::end() const {
  return index->end();
}

void Cluster::clear()
{
  index->clear();
  radius = 0;
}

void Cluster::build_ftable(VectString *data) {
  for(ContCluster::const_iterator i=index->begin(); i!=index->end(); i++) {
    ftable->insert(SimVect((*data)[pivot], (*data)[*i]));
  }
}

bool Cluster::improve(VectString& data) 
{
  SimType min_sum = 0;
  unsigned new_pivot = 0;
  for (ContCluster::const_iterator i = index->begin(); i != index->end(); ++i) 
    {
      SimType crt_sum = 0;
      for (ContCluster::const_iterator j = index->begin(); j != index->end(); ++j)
	{
	  crt_sum += SimDist(data[*i], data[*j]);
	}
      if (i == index->begin() || crt_sum < min_sum) 
	{
	  min_sum = crt_sum;
	  new_pivot = *i;
	}
    }
  if (new_pivot != pivot) 
    {
      pivot = new_pivot;
      return true;
    }
  return false;
}

ostream& operator << (ostream& out, const Cluster& c) {
  out << static_cast<unsigned>(c.index->size()) << endl;
  copy(c.index->begin(), c.index->end(), ostream_iterator<unsigned> (out, "\n"));
  out << endl << c.pivot << endl << endl;
  out << *c.ftable;
  out << endl << fixed << c.radius << endl;
  return out;
}

istream& operator>>(istream& in, Cluster& c) {
  /* cluster input:
     - n: number of elements
     - \n
     - v: elements
     - \n
     - p: pivot
     - \n
     - ftable
     - \n
     - r: radius
     - \n
  */
  unsigned n;
  in>>n;
  //return in;
  for (unsigned i=0; i<n; i++) {
    unsigned v;
    in>>v;
    c.index->insert(v);
    //if (i==10) return in;
  }
  in>>c.pivot;
  in >> *c.ftable;
  in>>c.radius;
  return in;
}

ContCluster::const_iterator Cluster::find(unsigned idx) const
{
  return index->find(idx);
}

void Cluster::insert(unsigned idx)
{
  index->insert(idx);
}

void Cluster::insert(unsigned idx, const SimVect vect)
{
  insert(idx);
  ftable->insert(vect);
}

void Cluster::erase(unsigned idx, const SimVect vect)
{
  index->erase(idx);
  ftable->erase(vect);
}
