#ifndef _Clusters_
// #define _Clusters_ 3
#define _Clusters_ Cluster_Method
// 1: Lexic
// 2: Medoids Ver 1
// 3: Medoids Ver 7

#include "../datastruct.h"
#include "../Sample.h"
#include "../PPDTable/PPDTable.h"
#include "Cluster.h"

extern unsigned CLUSTER_NO;

class QueryVect
{
public:
  unsigned query;
  SimVect vect;
  QueryVect()
  {
  }
  QueryVect(unsigned query, SimVect vect): query(query), vect(vect)
  {
  }

  friend ostream& operator<<(ostream& out, const QueryVect& qv)
  {
    return out << qv.query << "\t" << qv.vect;
  }
  friend istream& operator>>(istream& in, QueryVect& qv)
  {
    return in >> qv.query >> qv.vect;
  }  
};

typedef multimap<unsigned, QueryVect> ContQueryPivot;

class Clusters 
{
protected:
  VectString *data;
  VectCluster *clusters;
  PPDTable *ppdtable;
  ContQueryPivot *query_pivot;

public:
  Clusters(VectString *d);
  Clusters(const Clusters& cs);
  virtual ~Clusters();

  Clusters& operator=(const Clusters& cs);

  const size_t size_cluster() const;
  const Cluster& get_cluster(unsigned i) const;
  PPDTable& get_ppdtable() const;
  ContQueryPivot& get_query_pivot() const;

  void push_back_cluster(const Cluster& c);

  const VectClusterIt begin_cluster() const;
  const VectClusterIt end_cluster() const;

  const ContPPDEntryIt begin_ppdtable() const;
  const ContPPDEntryIt end_ppdtable() const;
  const ContPPDEntryIt find_ppdtable(const PPDEntry& e) const;
  const ContPPDEntryIt lower_bound_ppdtable(const PPDEntry& e) const;
  const ContPPDEntryIt upper_bound_ppdtable(const PPDEntry& e) const;

  void build_ftables();
  void build_ppdtable();

  bool is_pivot(unsigned idx) const;
  void insert(unsigned idx);
  void erase(unsigned idx);

  void hist(ostream &out, const VectString& data) const;
  void stat(ostream &out, const VectString& data) const;
  static ostream& info(ostream& out);

  friend ostream& operator<<(ostream& out, const Clusters& cs);
  friend istream& operator>>(istream& in, Clusters& cs);
};

ostream& operator<<(ostream& out, const ContQueryPivot& cont);
istream& operator>>(istream& in, ContQueryPivot& cont);

#endif
