#include "Lexic.h"

class Lexic_cmp
{
private:
  const VectString *data;
public:
  Lexic_cmp(const VectString *d): data(d) 
  {
  }
  bool operator()(unsigned a, unsigned b) const
  {
    return (*data)[a] < (*data)[b];
  }
};  

Lexic::Lexic(VectString *d): Clusters(d) {
  clusters=new VectCluster(CLUSTER_NO);
  ppdtable=new PPDTable();
	
  VectUnsigned d_id(d->size());
  for (unsigned i = 0; i < d->size(); i++) d_id[i] = i;

  sort(d_id.begin(), d_id.end(), Lexic_cmp(d));

  const unsigned n = static_cast<unsigned>(d->size());
  const unsigned cluster_no = CLUSTER_NO;
  const unsigned nclust = n/cluster_no;
  for (unsigned i = 0; i < cluster_no; i++) {
    const unsigned start = i*nclust;
    const unsigned end = (i == cluster_no? n:(i + 1) * nclust);
    const unsigned p = start + (end - start)/2;
    (*clusters)[i].set_pivot(d_id[p]);
    SimType radius = (*clusters)[i].get_radius();
    for (unsigned j = start; j < end; j++) {
      (*clusters)[i].insert(d_id[j]);
      radius = max(radius, SimDist((*data)[d_id[p]], (*data)[d_id[j]]));
    }
    (*clusters)[i].set_radius(radius);
  }
}

ostream& Lexic::info(ostream& out)
{
  Clusters::info(out);
  out << "Cluster method\tLexic" << endl;
  return out << endl;
}
