#include <iostream>

using namespace std;

#include "Medoids.h"

Medoids::Medoids(VectString* d): Clusters(d) 
{
  SAMPLE_SIZE  = SAMPLE_SIZE_MUL * CLUSTER_NO;
  if (SAMPLE_SIZE > d->size()) 
    {
      SAMPLE_SIZE = d->size();
    }

  clusters=new VectCluster(CLUSTER_NO);
  ppdtable=new PPDTable();

  time_t ltime;
  time(&ltime);
  srand(static_cast<unsigned int>(ltime));	

#if _Medoids_Ver_ == 1
  build_clusters_ver1();
#elif _Medoids_Ver_ == 2
  build_clusters_ver2();
#elif _Medoids_Ver_ == 3
  build_clusters_ver3();
#elif _Medoids_Ver_ == 4
  (*clusters)[0].set_pivot(1);
  (*clusters)[1].set_pivot(5);
  unsigned cluster_i=0;
  (*clusters)[cluster_i].push_back_index(0);
  (*clusters)[cluster_i].push_back_index(1);
  (*clusters)[cluster_i].push_back_index(4);
  cluster_i=1;
  (*clusters)[cluster_i].push_back_index(2);
  (*clusters)[cluster_i].push_back_index(3);
  (*clusters)[cluster_i].push_back_index(5);
#elif _Medoids_Ver_ == 5
  build_clusters_ver5();
#elif _Medoids_Ver_ == 6
  build_clusters_ver6();
#elif _Medoids_Ver_ == 7
  build_clusters_ver7();
#endif
}

void Medoids::build_clusters_ver1() 
{
  const unsigned data_size=static_cast<unsigned>(data->size());
  const unsigned cluster_no=CLUSTER_NO;
  const unsigned sample_size=SAMPLE_SIZE;
  const unsigned sample_sets=SAMPLE_SETS;
  const unsigned sample_steps=SAMPLE_STEPS;

  unsigned *pivot=new unsigned[cluster_no];
  unsigned *pivot_final=new unsigned[cluster_no];
  unsigned *sample=new unsigned[sample_size];

  SetUnsigned *id_set=new SetUnsigned;
  SetUnsigned *dataSet=new SetUnsigned();

  SimType min_sum_dist=0;

  for (unsigned l=0; l<sample_sets; l++) {
    // ---=== SAMPLE ===---
    // select sample records
    for (unsigned i=0; i<data_size; i++) dataSet->insert(i);
    for (unsigned i=0; i<sample_size; i++) {
      SetUnsignedIt it=dataSet->lower_bound(rand()%static_cast<unsigned>(dataSet->size()));
      unsigned id=*it;
      sample[i]=id;
      dataSet->erase(it);
    }

    for (unsigned k=0; k<sample_steps; k++){
      // ---=== BUILD ===---
      // select random pivots
      id_set->clear();
      for (unsigned i=0; id_set->size()<cluster_no;) { // BUILD (select pivots)
	unsigned j=sample[rand()%sample_size];
	if (id_set->insert(j).second)
	  pivot[i++]=j;
      }
		
      // assign each sample element to a pivot (to a cluster) and
      // compute the sum of dissimilarities  (sum of objective functions)
      SimType sum_dist=0;
      for (unsigned i=0; i<sample_size; i++) {
	// find the best pivot
	SimType min_dist=0;
	for (unsigned j=0; j<cluster_no; j++) {
	  SimType dist=SimDist((*data)[sample[i]], (*data)[pivot[j]]);
	  if (j==0 || dist<min_dist)
	    min_dist=dist;
	}
	sum_dist+=min_dist;
      }

      if ((l==0 && k==0) || sum_dist<min_sum_dist) {
	min_sum_dist=sum_dist;
	unsigned *pivot_extra;
	pivot_extra=pivot_final; pivot_final=pivot; pivot=pivot_extra;
      }
    }
  }

  delete dataSet;
  delete id_set;
  delete [] sample;

  // set final pivots as pivots
  for (unsigned i=0; i<cluster_no; i++)
    (*clusters)[i].set_pivot(pivot_final[i]);

  delete [] pivot_final;
  delete [] pivot;

  assign_clusters();
}

void Medoids::build_clusters_ver2() 
{
  const unsigned data_size=static_cast<unsigned>(data->size());
  const unsigned cluster_no=CLUSTER_NO;
  const unsigned sample_steps=SAMPLE_STEPS;

  unsigned *pivot=new unsigned[cluster_no];
  unsigned *pivot_final=new unsigned[cluster_no];

  SetUnsigned *id_set=new SetUnsigned;

  SimType min_sum_dist=0;
  for (unsigned k=0; k<sample_steps; k++){
    // ---=== BUILD ===---
    // select random pivots
    id_set->clear();
    for (unsigned i=0; id_set->size()<cluster_no;) { // BUILD (select pivots)
      unsigned j=rand()%data_size;
      if (id_set->insert(j).second)
	pivot[i++]=j;
    }
	
    // assign each sample element to a pivot (to a cluster) and
    // compute the sum of dissimilarities  (sum of objective functions)
    SimType sum_dist=0;
    for (unsigned i=0; i<data_size; i++) {
      // find the best pivot
      SimType min_dist=0;
      for (unsigned j=0; j<cluster_no; j++) {
	SimType dist=SimDist((*data)[i], (*data)[pivot[j]]);
	if (j==0 || dist<min_dist)
	  min_dist=dist;
      }
      sum_dist+=min_dist;
    }

    if (k==0 || sum_dist<min_sum_dist) {
      min_sum_dist=sum_dist;
      unsigned *pivot_extra;
      pivot_extra=pivot_final; pivot_final=pivot; pivot=pivot_extra;
    }
  }

  delete id_set;

  // set final pivots as pivots
  for (unsigned i=0; i<cluster_no; i++)
    (*clusters)[i].set_pivot(pivot_final[i]);

  delete [] pivot_final;
  delete [] pivot;

  assign_clusters();
}

void Medoids::build_clusters_ver3() 
{
  const unsigned data_size=static_cast<unsigned>(data->size());
  const unsigned cluster_no=CLUSTER_NO;
  const unsigned sample_steps=SAMPLE_STEPS;
  const unsigned sample_size=SAMPLE_SIZE;

  unsigned *pivot=new unsigned[cluster_no];
  unsigned *pivot_final=new unsigned[cluster_no];
  unsigned *sample=new unsigned[sample_size];

  SetUnsigned *id_set=new SetUnsigned;

  SimType min_sum_dist=0;
  for (unsigned k=0; k<sample_steps; k++){
    cout<<"k "<<k<<endl;
    // ---=== SAMPLE ===---
    // select sample records
    id_set->clear();

    if (k!=0) { // we already have a set of pivot_final
      // add the set of pivot_final to the sample
      for (unsigned i=0; i<cluster_no; i++) {
	id_set->insert(pivot_final[i]);
	sample[i]=pivot_final[i];
      }
    }

    for (unsigned i=id_set->size(); id_set->size()<sample_size;) { // 1st SAMPLE
      unsigned j=rand()%data_size;
      if (id_set->insert(j).second)
	sample[i++]=j;
    }

    cout<<"build start..."<<endl;
    // ---=== BUILD ===---	
//     id_set->clear();
//     for (unsigned i=0; i<sample_size; i++) id_set->insert(sample[i]);

    // first pivot
    unsigned pivot_id=0;
    SimType min_sum_dist_pivot=0;
    for (SetUnsignedIt i=id_set->begin(); i!=id_set->end(); i++) {
      SimType sum_dist=0;
      for (SetUnsignedIt j=id_set->begin(); j!=id_set->end(); j++)
	sum_dist+=SimDist((*data)[*i], (*data)[*j]);
      if (i==id_set->begin() || sum_dist<min_sum_dist_pivot) {
	pivot_id=*i;
	min_sum_dist_pivot=sum_dist;
      }
    }
    pivot[0]=pivot_id; // we have the first
    id_set->erase(pivot_id);
    cout << "first pivot..." << endl;

    // rest of the pivots
    for (unsigned m=1; m<cluster_no; m++) {
      SimType max_sum_Cji=0;
      unsigned pivot_id=0; // the best unselected object - the new pivot

      for (SetUnsignedIt i=id_set->begin(); i!=id_set->end(); i++) {
	SimType sum_Cji=0;
	for (SetUnsignedIt j=id_set->begin(); j!=id_set->end(); j++) {
	  // compute Dj
	  SimType Dj=0;
	  for (unsigned l=0; l<m; l++) {
	    SimType dist=SimDist((*data)[*j], (*data)[pivot[l]]);
	    if (l==0 || dist<Dj) {
	      Dj=dist;
	    }
	  }

	  // compute Cji
	  SimType dist=SimDist((*data)[*j], (*data)[*i]);
	  SimType Cji=Dj>dist?Dj-dist:0;

	  // add to total
	  sum_Cji+=Cji;
	}

	if (i==id_set->begin() || sum_Cji>max_sum_Cji) {
	  max_sum_Cji=sum_Cji;
	  pivot_id=*i;
	}
      }

      pivot[m]=pivot_id;
      id_set->erase(pivot_id);
    }
    cout<<"build end..."<<endl;

    // assign each sample element to a pivot (to a cluster) and
    // compute the sum of dissimilarities  (sum of objective functions)
    SimType sum_dist=0;
    for (unsigned i=0; i<sample_size; i++) {
      // find the best pivot
      SimType min_dist=0;
      for (unsigned j=0; j<cluster_no; j++) {
	SimType dist=SimDist((*data)[sample[i]], (*data)[pivot[j]]);
	if (j==0 || dist<min_dist)
	  min_dist=dist;
      }
      sum_dist+=min_dist;
    }

    if (k==0 || sum_dist<min_sum_dist) {
      min_sum_dist=sum_dist;
      unsigned *pivot_extra;
      pivot_extra=pivot_final; pivot_final=pivot; pivot=pivot_extra;
    }
  }

  delete id_set;
  delete [] sample;

  // set final pivots as pivots
  for (unsigned i=0; i<cluster_no; i++)
    (*clusters)[i].set_pivot(pivot_final[i]);

  delete [] pivot_final;
  delete [] pivot;
	
  assign_clusters();	
}

void Medoids::build_clusters_ver5() 
{ 
  const unsigned data_size = data->size();
  cout << "first pivot..." << endl;
  VectUnsigned *pivots = new VectUnsigned(CLUSTER_NO);
  SimType sum_min = 0;
  for (unsigned i = 0; i < data_size; i++) 
    {
      SimType sum_crt = 0;
      for (unsigned j = 0; j < data_size; j++) 
	{
	  if (i == j) continue;
	  sum_crt += SimDist((*data)[i], (*data)[j]);
	}
      if (i == 0 || sum_crt < sum_min) 
	{
	  sum_min = sum_crt;
	  (*pivots)[0] = i;
	}
    }
  VectUnsigned *assigned = new VectUnsigned();
  assigned->reserve(data_size);
  VectSimType *distances = new VectSimType();
  distances->reserve(data_size);
  for (unsigned i = 0; i < data_size; i++) {
    (*assigned)[i] = 0;
    (*distances)[i] = SimDist((*data)[(*pivots)[0]], (*data)[i]);
  }

  for (unsigned i = 1; i < CLUSTER_NO; i++) 
    {
      cout << "pivot[" << i << "]" << endl;
      SimType max_gain = 0;
      for (unsigned j = 0; j < data_size; j++) 
	{
	  unsigned k;
	  for (k = 0; k < i && (*pivots)[k] != j; k++);
	  if (k != i) continue;

	  SimType crt_gain = 0;
	  for (k = 0; k < data_size; k++) 
	    {
	      SimType crt_dist = SimDist((*data)[j], (*data)[k]);
	      if (crt_dist < (*distances)[k]) crt_gain += (*distances)[k] - crt_dist;
	    }
	  if (crt_gain > max_gain) 
	    {
	      max_gain = crt_gain;
	      (*pivots)[i] = j;
	    }
	}
      for (unsigned j = 0; j < data_size; j++)
	{
	  SimType new_dist = SimDist((*data)[(*pivots)[i]], (*data)[j]);
	  if (new_dist < (*distances)[j]) 
	    {
	      (*assigned)[j] = i;
	      (*distances)[j] = new_dist;
	    }
	}
    }
  
  cout << "finishing..." << endl;
  for (unsigned i = 0; i < CLUSTER_NO; i++) 
    {
      (*clusters)[i].set_pivot((*pivots)[i]);
    }
  for (unsigned i = 0; i < data_size; i++)
    {
      unsigned cluster_id = (*assigned)[i];
      (*clusters)[cluster_id].insert(i);
      (*clusters)[cluster_id].set_radius(max((*clusters)[cluster_id].get_radius(),
					     (*distances)[i]));
    }
  delete distances;
  delete assigned;
  delete pivots;
}

void Medoids::build_clusters_ver6() 
{
  const unsigned data_size = data->size();
  SetUnsigned *pivots = new SetUnsigned();

  time_t ltime;
  time(&ltime);
  srand(static_cast<unsigned>(ltime));

  while (pivots->size() < CLUSTER_NO) pivots->insert(rand() % data_size);
  
  unsigned i = 0;
  for (SetUnsignedIt it = pivots->begin(); it != pivots->end(); ++it) (*clusters)[i++].set_pivot(*it);

  delete pivots;

  assign_clusters();

  for (unsigned i = 0; i < SAMPLE_STEPS; i++) 
    {
      bool changed = false;
      for (VectClusterIt it = clusters->begin(); it != clusters->end(); ++it) 
	{
	  if (it->improve(*data)) changed = true;
	}

      if (changed) 
	{
	  cout << "imporved (" << (i + 1) << "/" << SAMPLE_STEPS << ")..." << endl;
	  for (VectClusterIt it = clusters->begin(); it != clusters->end(); ++it) it->clear();
	  assign_clusters();
	}
      else break;
    }
}

void Medoids::build_clusters_ver7() 
{
  const unsigned data_size = data->size();

  time_t ltime;
  time(&ltime);
  srand(static_cast<unsigned>(ltime));

  Sample
    samples = Sample(SAMPLE_SIZE, data_size),
    pivots = Sample(CLUSTER_NO, SAMPLE_SIZE);
  
  unsigned i = 0;
  for (vector<unsigned>::const_iterator it = pivots.begin(); it != pivots.end(); ++it)
    (*clusters)[i++].set_pivot(samples[*it]);

  assign_clusters(samples.begin(), samples.end());

  for (unsigned i = 0; i < SAMPLE_STEPS; i++) 
    {
      bool changed = false;
      for (VectClusterIt it = clusters->begin(); it != clusters->end(); ++it) 
	{
	  if (it->improve(*data)) changed = true;
	}

      if (changed) 
	{
	  cerr << "...imporve clusters\t" << (i + 1) << "/" << SAMPLE_STEPS << endl;
	  for (VectClusterIt it = clusters->begin(); it != clusters->end(); ++it) it->clear();
	  assign_clusters(samples.begin(), samples.end());
	}
      else break;
    }

  cerr << "...assign clusters" << endl;
  assign_clusters_minus(samples.begin(), samples.end());
}

void Medoids::assign_clusters() 
{
  const unsigned data_size = data->size();

  // assign each data element to a pivot (to a cluster)
  for (unsigned i=0; i<data_size; i++) {
    SimType min_dist=0;
    unsigned cluster_i=0;
		
    // find the best pivot
    for (unsigned j = 0; j < CLUSTER_NO; j++) {
      if (i == (*clusters)[j].get_pivot()) {
	min_dist=0;
	cluster_i=j;
	break;
      }
      SimType dist = SimDist((*data)[i], (*data)[(*clusters)[j].get_pivot()]);
      if (j == 0 || dist < min_dist) {
	min_dist = dist;
	cluster_i = j;
      }	
    }

    (*clusters)[cluster_i].insert(i);
    (*clusters)[cluster_i].set_radius(max((*clusters)[cluster_i].get_radius(), min_dist));
  }
}

void Medoids::assign_clusters(vector<unsigned>::const_iterator begin, vector<unsigned>::const_iterator end)
{
  // assign each data element to a pivot (to a cluster)
  for (vector<unsigned>::const_iterator i = begin; i != end; ++i) {
    SimType min_dist=0;
    unsigned cluster_i=0;
		
    // find the best pivot
    for (unsigned j = 0; j < CLUSTER_NO; j++) {
      if (*i == (*clusters)[j].get_pivot()) {
	min_dist=0;
	cluster_i=j;
	break;
      }
      SimType dist = SimDist((*data)[*i], (*data)[(*clusters)[j].get_pivot()]);
      if (j == 0 || dist < min_dist) {
	min_dist = dist;
	cluster_i = j;
      }	
    }

    (*clusters)[cluster_i].insert(*i);
    (*clusters)[cluster_i].set_radius(max((*clusters)[cluster_i].get_radius(), min_dist));
  }
}

void Medoids::assign_clusters_minus(vector<unsigned>::const_iterator begin, vector<unsigned>::const_iterator end)
{
  SetUnsigned samples = SetUnsigned(begin, end);

  const unsigned data_size = data->size();

  // assign each data element to a pivot (to a cluster)
  for (unsigned i=0; i<data_size; i++) {
    if (samples.find(i) != samples.end()) continue;

    SimType min_dist=0;
    unsigned cluster_i=0;
		
    // find the best pivot
    for (unsigned j = 0; j < CLUSTER_NO; j++) {
      if (i == (*clusters)[j].get_pivot()) {
	min_dist=0;
	cluster_i=j;
	break;
      }
      SimType dist = SimDist((*data)[i], (*data)[(*clusters)[j].get_pivot()]);
      if (j == 0 || dist < min_dist) {
	min_dist = dist;
	cluster_i = j;
      }	
    }

    (*clusters)[cluster_i].insert(i);
    (*clusters)[cluster_i].set_radius(max((*clusters)[cluster_i].get_radius(), min_dist));
  }
}

ostream& Medoids::info(ostream& out)
{
  Clusters::info(out);
  out << "Cluster method\tMedoids " << _Medoids_Ver_ << endl;
  out << "sample size\t" << SAMPLE_SIZE_MUL << endl;
  out << "sample sets\t" << SAMPLE_SETS << endl;
  out << "sample steps\t" << SAMPLE_STEPS << endl;
  return out << endl;
}

ostream& output(ostream& out, const unsigned* p, const unsigned k, const VectString& d)
{
  for (unsigned i=0; i<k; i++)
    out<<'['<<i<<"]\t"<<p[i]<<"\t("<<d[p[i]]<<')'<<endl;
  return out;
}

ostream& output(ostream& out, const Closest* c, const unsigned n_k, const unsigned* p, const unsigned* np) 
{
  for (unsigned i=0; i<n_k; i++)
    out<<np[i]<<":\t("<<p[c[i].d]<<','<<p[c[i].e]<<')'<<endl;
  return out;
}
