#ifndef _Medoids_
#define _Medoids_

#include <iostream>
#include <vector>

using namespace std;

#include "Clusters.h"

const unsigned SAMPLE_SIZE_MUL = 5;
const unsigned SAMPLE_SETS     = 10;
const unsigned SAMPLE_STEPS    = 10; // sample steps for each sample

#if   _Clusters_ == 2
#define _Medoids_Ver_ 1
#elif _Clusters_ == 3
#define _Medoids_Ver_ 7
#else
#define _Medoids_Ver_ -1
#endif
// 1: random sample, random pivots
// 2: sample is the full data
// 3: swap step is used
// 4: predifined clusters
// 5: PAM: Partitioning Around Medoids; k-Medoids algorithm 
// 6: ML - Medoids
// 7: ML - Medoids, Sample

class Medoids : public Clusters {
public:
  Medoids(VectString* d);
  static ostream& info(ostream& out);

private:
  unsigned SAMPLE_SIZE;

  void build_clusters_ver1();
  void build_clusters_ver2();
  void build_clusters_ver3();
  // ver4 is inside Medoids()
  void build_clusters_ver5();
  void build_clusters_ver6();
  void build_clusters_ver7();
  void assign_clusters();
  void assign_clusters(vector<unsigned>::const_iterator begin, vector<unsigned>::const_iterator end);
  void assign_clusters_minus(vector<unsigned>::const_iterator begin, vector<unsigned>::const_iterator end);
};

typedef struct {
  unsigned int d, e;
} Closest;

typedef vector<Closest> VectClosest;
typedef VectClosest::iterator VectClosestIt;

ostream& output(ostream& out, const unsigned* p, const unsigned k, const VectString& d);
ostream& output(ostream& out, const Closest* c, const unsigned n_k, const unsigned* p, const unsigned* np);

#endif
