#include "FreqEst.h"

unsigned lookup = 0, 
  lookup_hit = 0, 
  lookup_miss = 0;

float FreqEstFunc(VectString& d, Clusters& cs, Predicate p) {
  float est = 0;
  float fract;

  for (VectClusterIt i=cs.begin_cluster(); i!=cs.end_cluster(); i++/*, k++*/) {
    float est_cluster=0;
    SimVect v1=SimVect(p.query, d[i->get_pivot()]);
    SimType v1d=v1.get_dist();
    // pruning
    if (v1d>i->get_radius()+p.dist)
      continue;
    for (ContFTable::const_iterator j=i->get_ftable().begin(); j!=i->get_ftable().end(); j++) {
      SimVect v2=j->first;
      SimType v2d=v2.get_dist();

      // pruning
      if (v1d + v2d <= p.dist) {
	est_cluster += j->second;
	continue;
      }

      // pruning
      if (v1d > p.dist + v2d || v2d > p.dist + v1d) {
	continue;
      }

      PPDEntry e=PPDEntry(v1, v2, p.dist, 0, 0);
      ContPPDEntryIt it_e=cs.find_ppdtable(e);

      lookup++;

      if (it_e==cs.end_ppdtable()) {
	it_e=cs.lower_bound_ppdtable(e);
	--it_e;
	// !!! FreqEst assumes that ContPPDEntry is an ordered container !!!
	if (it_e->vect1!=v1 || it_e->vect2!=v2) {
	  // !!! it_e==cs.end_ppdtable() || !!! || it_e->dist_m>p.dist !!!

	  lookup_miss++;

	  continue;
	}				
      }

      lookup_hit++;
 
#ifndef _PPD_online_
      fract = it_e->fract;
#endif

#ifdef _PPD_online_
      ContPPDEntryIt it_e_next = it_e;
      it_e_next++;
      while (it_e->vect1 == it_e_next->vect1 && it_e->vect2 == it_e_next->vect2) {
	it_e_next++;
      }
      it_e_next--;

      fract = static_cast<float>(it_e->count_e) / it_e_next->count_e;
#endif

      est_cluster += fract * j->second;
    }
    est+=est_cluster;
  }

  return est;
}

unsigned FreqRealFunc(VectString& d, Predicate p, VectString & r) {
  for (VectStringIt i=d.begin(); i!=d.end(); i++) {
    if (SimDist(p.query, *i)<=p.dist) {
      r.push_back(*i);
    }
  }
  return r.size();
}

unsigned FreqRealFunc(VectString& d, Predicate p) {
  VectString r;
  return FreqRealFunc(d, p, r);
}  

float FreqPPDFunc(VectString& d, Clusters& cs, Predicate p)
{
  float count = 0;

  for (VectClusterIt i=cs.begin_cluster(); i!=cs.end_cluster(); i++/*, k++*/) {
    SimVect v1=SimVect(p.query, d[i->get_pivot()]);

    for (ContCluster::const_iterator j = i->begin(); j != i->end(); j++) {
//       // pruning
//       SimType dist = SimDist(p.query, d[*j]);
//       if (dist > p.dist) {
// 	continue;
//       }

      SimVect v2=SimVect(d[i->get_pivot()], d[*j]);
      PPDEntry e=PPDEntry(v1, v2, p.dist, 0, 0);

      ContPPDEntryIt it_e=cs.find_ppdtable(e);
      if (it_e==cs.end_ppdtable()) {
	it_e=cs.lower_bound_ppdtable(e); 
	--it_e;
	// !!! FreqEst assumes that ContPPDEntry is an ordered container !!!
	if (it_e->vect1!=v1 || it_e->vect2!=v2) {
	  // !!! it_e==cs.end_ppdtable() || !!! || it_e->dist_m>p.dist !!!
	  continue;
	}				
      }
      count += it_e->fract;
    }
  }
  return count;
}
