#ifndef _Record_
#define _Record_

#include <iostream>
#include <iterator>
#include <vector>

using namespace std;

#include "Predicate.h"

class Record: public Predicate 
{
public:
  float est, rel_err;
  unsigned real;
  unsigned cnt_cluster, cnt_ppdpair;

  Record();
  Record(icstring q, SimType d);
  Record(icstring q, SimType d, float e, unsigned r);
  Record(icstring q, SimType d, float e, unsigned r, unsigned cnt_c, unsigned cnt_p);
  Record(Predicate p, float e, unsigned r);
  Record(Predicate p, float e, unsigned r, unsigned cnt_c, unsigned cnt_p);

  void setRelErr();

  ostream& toWeka(ostream& out);
};

typedef vector<Record> VectRecord;
typedef VectRecord::iterator VectRecordIt;

inline ostream& operator<<(ostream& out, const VectRecord& v) 
{
  out<<static_cast<unsigned>(v.size())<<endl<<endl;
  copy(v.begin(), v.end(), ostream_iterator<Record>(out, "\n"));
  return out;
}

istream& operator>>(istream& in, VectRecord& v);

ostream & VectRecord_toWeka(ostream & out,  VectRecord & v);

inline Record::Record(): Predicate(), est(0), rel_err(-1), real(0) 
{
}

inline Record::Record(icstring q, SimType d): Predicate(q, d), est(0), rel_err(-1), real(0) 
{
}

inline Record::Record(icstring q, SimType d, float e, unsigned r): Predicate(q, d), est(e), real(r) 
{
  setRelErr();
}

inline Record::Record(icstring q, SimType d, float e, unsigned r, unsigned cnt_c, unsigned cnt_p): 
  Predicate(q, d), est(e), real(r), cnt_cluster(cnt_c), cnt_ppdpair(cnt_p) 
{
  setRelErr();
}

inline Record::Record(Predicate p, float e, unsigned r): Predicate(p), est(e), real(r) 
{
  setRelErr();
}

inline Record::Record(Predicate p, float e, unsigned r, unsigned cnt_c, unsigned cnt_p):
  Predicate(p), est(e), real(r), cnt_cluster(cnt_c), cnt_ppdpair(cnt_p) 
{
  setRelErr();
}

inline void Record::setRelErr() 
{
  if (real!=0) {
    rel_err=static_cast<float>(est-real)/real;
//     if (rel_err<0) {
//       rel_err-=rel_err;
//     }
  }
  else {
    rel_err=-1;
  }
} 

inline ostream& Record::toWeka(ostream & out) 
{
  out << query.length() << "," << dist << "," << est;
  out << "," << cnt_cluster << "," << cnt_ppdpair;
  out << "," << rel_err;
  return out;
}

inline ostream& operator<<(ostream& out, const Record& r) 
{
  out << r.query << "\t" << r.dist << "\t" << r.real << "\t" << r.est << "\t" << r.rel_err;
  out << "\t" << r.cnt_cluster << "\t" << r.cnt_ppdpair;
  return out;
}

inline istream& operator>>(istream& in, Record& r) 
{
  in >> r.query >> r.dist >> r.real >> r.est >> r.rel_err;
  in >> r.cnt_cluster >> r.cnt_ppdpair;
  return in;
}

inline float Record_getEst(Record r) 
{
  return r.est;
}

inline float Record_getLen(Record r) 
{
  return static_cast<float>(r.query.length());
}

inline float Record_getThr(Record r) 
{
  return static_cast<float>(r.dist);
}

typedef float Record_Func(Record);

#endif
