#include "Test.h"
#include "FreqEst.h"

#include <cmath>
#include <sys/times.h>
#include <unistd.h>

using namespace std;

Test::Test(VectString* dat, Clusters* clust, bool doBuck):
  data(dat), clusters(clust), errorCorr(NULL), doBuck(doBuck)
{
  if (doBuck) 
    {
      for (unsigned i = 0; i < 10; i++)
        {
          buck[i] = 0;
        }
    }
}

void Test::create(const SetUnsigned& setQuery) 
{
  cerr << "...build test" << endl;
  for (SetUnsignedCIt it=setQuery.begin(); it!=setQuery.end(); it++) {
    SimType dist;

#if _SimDist_ == 1
    dist = static_cast<SimType>(rand() % (THRESHOLD_MAX - THRESHOLD_MIN + 1) + THRESHOLD_MIN);
#elif _SimDist_ == 2
    dist = static_cast<SimType>(rand() % static_cast<int>(THRESHOLD_MAX * 10 - THRESHOLD_MIN * 10 + 1)) / 10 + THRESHOLD_MIN;
#endif

    Predicate p = Predicate((*data)[*it], dist);
    p.freq = FreqRealFunc(*data, p);

    vectTest.push_back(p);
  }
}

void Test::test() 
{
  cerr << "...test" << endl;

  sum_rel_err=0;
  sum_rel_err_cor=0;
  no_rel_err = 0;

  sum_time = 0;
  clock_t start_time, end_time;
  struct tms buff;

  for (unsigned i = 0; i < vectTest.size(); i++)
    {
      Predicate p=vectTest[i];
      float f,  f_cor;
      unsigned r;

      start_time = times(&buff);
      f = FreqEstFunc(*data, *clusters, p);      
      end_time = times(&buff);      
      sum_time += static_cast<float>(end_time - start_time) / sysconf(_SC_CLK_TCK);
      r = p.freq;

      if (r != 0) {
        float  rel_err = (f - r) / r;
        sum_rel_err += fabs(rel_err);
	
        if (errorCorr)
          {
            float r_cor = errorCorr->getError(Record(p, f, r));

            if (r_cor == -1)
              {
                f_cor = f * 2;
              }
            else
              {
                f_cor = f / (r_cor + 1);
              }

            float rel_err_cor = (f_cor - r) / r;

            sum_rel_err_cor += fabs(rel_err_cor);

            // buckets
            if (doBuck)
              {
                int idx = static_cast<int>(rel_err_cor * 4) + 5;
                buck[idx]++;
              }
          }

        no_rel_err++;

        // 	// adjust 1
        // 	if (r>=REAL_THRESHOLD) { // adjusting
        // 	  sum_rel_err_adj+=rel_err;
        // 	  sum_rel_err_cor_adj+=rel_err_cor;
        // 	  no_rel_err_adj++;
        // 	}

        // 	// adjunst 2
        // 	unsigned r_adj2 = r < REAL_THRESHOLD ? REAL_THRESHOLD : r;
        // 	sum_rel_err_adj2 += fabs(f - r_adj2) / r_adj2;
        // 	sum_rel_err_cor_adj2 += fabs(f_cor - r_adj2) / r_adj2;
	
        // 	// adjust 3
        // 	sum_rel_err_adj3 += r < REAL_THRESHOLD ? 0 : rel_err;
        // 	sum_rel_err_cor_adj3 += r < REAL_THRESHOLD ? 0 : rel_err_cor;

      }
      else {
        cerr << "...absolute error" << endl;
      }

      // 	cout << i << "\t" << setw(20) << p.query << "\t(" << p.freq << ")\t" << f << endl;
    }
}

ostream& Test::info(ostream& out) const
{
  out << "Test" << endl << "---" << endl;
  out << "test size\t" << vectTest.size() << endl;
  out << "threshold\t" << THRESHOLD_MIN << " " << THRESHOLD_MAX << endl;
  out << endl;
  
  int relative_error = static_cast<int>((sum_rel_err/no_rel_err) * 100);
  int relative_error_corrected = static_cast<int>((sum_rel_err_cor/no_rel_err) * 100);
  float estimation_time = sum_time / vectTest.size();

  out << "Relative error\t" << relative_error
      << "%" << endl;
  if (errorCorr) 
    {
      out << "Relative error corrected\t" << relative_error_corrected
          << "%" << endl;
      out << "Relative error improvement\t" << relative_error - relative_error_corrected
          << "%" << endl;
    }
  out << "Estimation time\t" << estimation_time << endl;

  if (doBuck)
    {
      out << "Relative error buckets\t10" << endl;
      for (int i = 0; i < 10; i++)
        {
          out << (-125 + i * 25) << "\t" << static_cast<float>(buck[i]) / vectTest.size() << endl;
        }
    }

  return out << endl;
}

void Test::setErrorCorr(const ErrorCorr& errorCorr) 
{
  this->errorCorr = &errorCorr;
}


