#include "CLOSE_UNIQUE.h"

#include <algorithm>
#include <iterator>

using namespace std;

CLOSE_UNIQUE::CLOSE_UNIQUE(VectString *d, Clusters *c, ContQueryPivot *q): PPDSample(d, c, q)
{
	cout << "---check CLOSE_UNIQUE---" << endl;

  // init - sample
  sample = Sample(data->size());
  sample_max = static_cast<unsigned>(count_unique() * .4);
  build_sample();

  // init - loop
  sample_idx = 0;
  unsigned j=samplePair.begin()->index;
  samplePair.erase(samplePair.begin());
  cluster=clusters->get_cluster(j);
  string_it=cluster.begin();
}

void CLOSE_UNIQUE::step() 
{
  string_it++;
  if (string_it==cluster.end()) {
    if (samplePair.empty()) {
      sample_idx++;
      if (sample_idx < max(1u, static_cast<unsigned>(SAMPLE_PERCENT * data->size())))
        build_sample();
      else
        return;
    }
    unsigned i=samplePair.begin()->index;
    samplePair.erase(samplePair.begin());
    cluster=clusters->get_cluster(i);
    string_it=cluster.begin();
  }
}

bool CLOSE_UNIQUE::has_next() const 
{
  if (sample_idx < sample_max && sample.has_next())
    return true;
  return false;
}

PPDTriple CLOSE_UNIQUE::next() 
{
  const unsigned q=sample_idx;
  const unsigned p=cluster.get_pivot();
  const unsigned s=*string_it;

  PPDTriple triple=PPDTriple(q,
                             p,
                             s,
                             SimVect((*data)[q], (*data)[p]),
                             SimVect((*data)[p], (*data)[s]),
                             SimDist((*data)[q], (*data)[s]));

  // increment
  step();

  return triple;
}

void CLOSE_UNIQUE::build_sample() 
{
  unsigned sample_crt = 0; 
  bool done = false;
  set<SimVect>
    newVect, 
    diffVect;

  while (!done) 
    {
      if (!sample.has_next()) {
        break;
      }
      sample_crt = sample.generate();
      samplePair.clear();
      newVect.clear();

      for (unsigned i = 0; i < clusters->size_cluster(); i++) {
        const unsigned p = clusters->get_cluster(i).get_pivot();
        SimVect vect = SimVect((*data)[sample_crt], (*data)[p]);
        if (samplePair.size() < QUEUE_SIZE || 
            vect.get_dist() < samplePair.begin()->vect.get_dist())
          {
            samplePair.insert(IndexEntryVect(i, vect));
            newVect.insert(vect);
            if (samplePair.size() > QUEUE_SIZE)
              {
                samplePair.erase(samplePair.begin());
                newVect.erase(vect);
              }
          }
      }

      diffVect.clear();
      set_difference(newVect.begin(), newVect.end(), 
                     uniqueVect.begin(), uniqueVect.end(), 
                     inserter(diffVect, diffVect.begin()));
      if (diffVect.size() >= UNIQUE_NO) 
        done = true;
    }
  
  sample.push_back(sample_crt);
  copy(diffVect.begin(), diffVect.end(), inserter(uniqueVect, uniqueVect.begin()));
}

unsigned CLOSE_UNIQUE::count_unique() 
{
  unsigned sample_crt = 0; 
  set<SimVect>
    diffVect;

  for (unsigned j = 0; j < data->size(); j++) 
    {      
      sample_crt = j;

      for (unsigned i = 0; i < clusters->size_cluster(); i++) {
        const unsigned p = clusters->get_cluster(i).get_pivot();
        SimVect vect = SimVect((*data)[sample_crt], (*data)[p]);

        diffVect.insert(vect);
	
      }
    }
  
  return diffVect.size();
}
