#ifndef _PPDEntry_
#define _PPDEntry_

#include <iomanip>
#include <iostream>
#include <iterator>
#include <set>

using namespace std;

#include "../SimFunc/SimVect.h"

#if _SimDist_ == 1
const unsigned DIST_THRESHOLD = 5;
#elif _SimDist_ == 2
const float DIST_THRESHOLD = static_cast<float>(0.96);
#endif

class PPDEntry {
public:
  SimVect vect1, vect2;
  SimType dist_m;
  unsigned count_e;
  float fract;

  PPDEntry();
  PPDEntry(SimVect v1, SimVect v2, SimType d, unsigned c = 1, float f = 0);

  operator size_t() const;

#ifdef _Sim_EditVect_
  void serialize(ofstream & out) const;
  void deserialize(ifstream & in);
#endif

  friend ostream& operator<<(ostream& out, const PPDEntry& e);
  friend istream& operator>>(istream& in, PPDEntry& e);
  friend bool operator<(const PPDEntry& left, const PPDEntry& right);
};

typedef set<PPDEntry> ContPPDEntry; // !!! FreqEst assumes that ContPPDEntry is an ordered container !!!
                                    // !!! PPDTable assumes that ContPPDEntry is an ordered container !!!
                                    // !!! PPDTable assumes that ContPPDEntry is a container where remove does not imvalidate iterators !!!
typedef ContPPDEntry::iterator ContPPDEntryIt;

ostream& operator<<(ostream& out, const ContPPDEntry& c);
istream& operator>>(istream& in, ContPPDEntry& c);

inline PPDEntry::PPDEntry(): dist_m(0), count_e(0), fract(0) {
}

inline PPDEntry::PPDEntry(SimVect v1, SimVect v2, SimType d, unsigned c, float f): 
  vect1(v1), vect2(v2), dist_m(d), count_e(c), fract(f)
{
  if (dist_m > DIST_THRESHOLD) 
    {
      dist_m = DIST_THRESHOLD;
    }
}

inline PPDEntry::operator size_t() const {
  // (size_t) SimVect is 3 digits
  // (size_t) SimType is 2 digits
  return static_cast<size_t>(vect1)*100000+static_cast<size_t>(vect2)*100+SimType2size_t(dist_m);
}

inline ostream& operator<<(ostream& out, const PPDEntry& e) {
  return out<<e.vect1<<"\t"<<e.vect2<<"\t"<<setw(5)<<e.dist_m<<"\t"<<setw(5)<<e.count_e<<"\t"<<e.fract;
}

inline istream& operator>>(istream& in, PPDEntry& e) {
  return in>>e.vect1>>e.vect2>>e.dist_m>>e.count_e>>e.fract;
}

inline bool operator<(const PPDEntry& left, const PPDEntry& right) {
  if (&left==&right)
    return false;
  if (left.vect1!=right.vect1)
    return left.vect1<right.vect1;
  if (left.vect2!=right.vect2)
    return left.vect2<right.vect2;
  return left.dist_m<right.dist_m;
}

inline ostream& operator<<(ostream& out, const ContPPDEntry& c) {
  out<<static_cast<unsigned>(c.size())<<endl<<endl;
  copy(c.begin(), c.end(), ostream_iterator<PPDEntry>(out, "\n"));
  return out;
}

#endif
