#include "PPDTable.h"

PPDTable::PPDTable() 
{
  table=new ContPPDEntry;
}

PPDTable::PPDTable(const PPDTable& t) 
{
  table=new ContPPDEntry;
  *table=*t.table;
}

PPDTable::~PPDTable() 
{
  delete table;
}

PPDTable& PPDTable::operator=(const PPDTable& t) 
{
  if (this==&t)
    return *this;
  *table=*t.table;
  return *this;
}

const size_t PPDTable::size_table() const 
{
  return table->size();
}

const ContPPDEntryIt PPDTable::begin_table() const 
{
  return table->begin();
}

const ContPPDEntryIt PPDTable::end_table() const 
{
  return table->end();
}

const ContPPDEntryIt PPDTable::find_table(const PPDEntry& e) const 
{
  return table->find(e);
}

const ContPPDEntryIt PPDTable::lower_bound_table(const PPDEntry& e) const 
{
  return table->lower_bound(e);
}

const ContPPDEntryIt PPDTable::upper_bound_table(const PPDEntry& e) const 
{
  return table->upper_bound(e);
}

ostream& operator<<(ostream& out, const PPDTable& t) 
{
  return out << *t.table;
}

istream& operator>>(istream& in, PPDTable& t) 
{
  return in >> *t.table;
}

void PPDTable::insert(const PPDEntry& e) 
{
  PPDEntry e_min=PPDEntry(e.vect1, e.vect2, SIM_TYPE_MIN, 0, 0);
  PPDEntry e_max=PPDEntry(e.vect1, e.vect2, SIM_TYPE_MAX, 0, 0);

  PPDEntry e_new=e; e_new.count_e=0;
  // the element that will be added

  ContPPDEntryIt it_min=table->lower_bound(e_min);
  ContPPDEntryIt it_max=table->upper_bound(e_max);

  // between it_min and it_max are all the elements with the same vect1 and vect2
  unsigned a=0;
  // "a" holds the total number of occurences of pair (v1, v2)
  if (it_min!=it_max) { // there is at lease one pair with (v1, v2)
    ContPPDEntryIt it_last=it_max; --it_last;
    a=it_last->count_e;

    ContPPDEntryIt it_prev=table->lower_bound(e); --it_prev; // !!! PPDTable assumes that ContPPDEntry is an ordered container !!!
    if (it_prev->vect1==e_new.vect1 && it_prev->vect2==e_new.vect2) // there is an entry (v1, v2, d) with d>=e_new.d
      e_new.count_e=it_prev->count_e;
  }

  // add new element (if necessary)
  if (table->find(e_new)==table->end()) {
    bool do_insert=false;
    if (e_new.dist_m>DIST_THRESHOLD)
      // if e_new.dist_m > it_last->dist_m
      //   insert e_new
      if (it_min!=it_max) {
	ContPPDEntryIt it_last=it_max; --it_last;
	if (e.dist_m>it_last->dist_m) {
	  table->erase(it_last);
	  do_insert=true;
	}
      } else
	do_insert=true;
    else
      do_insert=true;
    if (do_insert) {
      table->insert(e_new);
      // if the element is added, the iterators need to be updated
      it_min=table->lower_bound(e_min);
      it_max=table->upper_bound(e_max);
    }
  }

  // update count_e and fract of all the elements with (vect1, vect2)
  a++;
  ContPPDEntry table_new;
  for (ContPPDEntryIt it=it_min; it!=it_max; ++it) {
    PPDEntry e_tmp=*it;
    if (e_tmp.dist_m >= e.dist_m)
      e_tmp.count_e+=e.count_e;
    e_tmp.fract=static_cast<float>(e_tmp.count_e)/a;
    table_new.insert(e_tmp);
  }
  table->erase(it_min, it_max);
  table->insert(table_new.begin(), table_new.end());
}

void PPDTable::erase(const PPDEntry& e) 
{
  ContPPDEntryIt it = table->find(e);
  if (it == table->end()) 
    {
      cerr << "PPDTable::erase entry not found\t" << e << endl;
      return;
    }

  if (it->count_e == 0)
    {
      cerr << "PPDTable::erase entry has cont 0\t" << e << endl;
      return;
    }

  PPDEntry e_min=PPDEntry(e.vect1, e.vect2, SIM_TYPE_MIN, 0, 0);
  PPDEntry e_max=PPDEntry(e.vect1, e.vect2, SIM_TYPE_MAX, 0, 0);

  ContPPDEntryIt it_min=table->lower_bound(e_min);
  ContPPDEntryIt it_max=table->upper_bound(e_max);

  // between it_min and it_max are all the elements with the same vect1 and vect2
  unsigned a=0;
  // "a" holds the total number of occurences of pair (v1, v2)
  if (it_min == it_max)
    {
      cerr << "PPDTable::erase entry range not found\t" << e << endl;
      return;
    }  
  // there is at lease one pair with (v1, v2)
  ContPPDEntryIt it_last=it_max; --it_last;
  a = it_last->count_e;

  // update count_e and fract of all the elements with (vect1, vect2)
  a--;
  ContPPDEntry table_new;
  for (ContPPDEntryIt it = it_min; it != it_max; ++it) {
    PPDEntry e_tmp = *it;
    if (e_tmp.dist_m >= e.dist_m)
      {
	e_tmp.count_e -= e.count_e;
      }
    e_tmp.fract = static_cast<float>(e_tmp.count_e) / a;
    if (e_tmp.count_e != 0) 
      {
	table_new.insert(e_tmp);
      }
  }
  table->erase(it_min, it_max);
  table->insert(table_new.begin(), table_new.end());
}

void PPDTable::prune_max() {
  for (ContPPDEntryIt it=table->begin(); it!=table->end();)
    if (it->dist_m > DIST_THRESHOLD) {
      ContPPDEntryIt er = it;
      it++;
      // !!! PPDTable assumes that ContPPDEntry is a container where remove does not invalidate iterators !!!
      table->erase(er);
    } else {
      it++;
    }
}

#ifdef _Sim_EditVect_

void PPDTable::serialize(ofstream &  out)
{
  unsigned sz = static_cast<unsigned>(table->size());
  out.write(reinterpret_cast<char *>(&sz), sizeof(unsigned));
  for(ContPPDEntryIt it = table->begin(); it != table->end(); it++) {
    it->serialize(out);
  }
}

void PPDTable::deserialize(ifstream &  in)
{
  unsigned sz;
  in.read(reinterpret_cast<char *>(&sz), sizeof(unsigned));
  for (unsigned i = 0; i < sz; i++) {
    PPDEntry e;
    e.deserialize(in);
    table->insert(e);
  }
}

#endif
