#include "SEPIA.h"
#include "misc.h"
#include "Util.h"
#include "Cluster/ClustersType.h"
#include "PPDTable/PPDSample.h"
#include "FreqEst/FreqEst.h"
#include "FreqEst/ErrorCorr.h"
#include "FreqEst/Update.h"
#include "SimFunc/EditDist.h"

unsigned DATASET_SIZE = 0;
unsigned UPDATES_SIZE = 0;

#if _SimDist_ == 1
extern unsigned THRESHOLD_MIN;
extern unsigned THRESHOLD_MAX;
#elif _SimDist_ == 2
extern float THRESHOLD_MIN;
extern float THRESHOLD_MAX;
#endif

int main(int argc, char* argv[])
{
  int no_arg = 8;
  if (DO_UPDATES) 
    {
      no_arg++;
    }
  if (argc != no_arg) 
    {
      cerr << argv[0] << " DATA_SIZE CLUSTER_NO SAMPLE_PERCENT SAMPLE_QUEUE THR_MIN THR_MAX";
      if (DO_UPDATES) 
        {
          cerr << " UPDATES_SIZE";
        }
      cerr << " FIRST_TIME" << endl;
      return 1;
    }

  DATASET_SIZE = atoi(argv[1]);
  CLUSTER_NO = atoi(argv[2]);
  SAMPLE_PERCENT = atoi(argv[3]) / 10000.;
  QUEUE_SIZE = atoi(argv[4]);
#if _SimDist_ == 1
  THRESHOLD_MIN  = atoi(argv[5]);
  THRESHOLD_MAX  = atoi(argv[6]);
#elif _SimDist_ == 2
  THRESHOLD_MIN     = atoi(argv[5]) / 100.;
  THRESHOLD_MAX     = atoi(argv[6]) / 100.;
#endif
#ifdef DO_UPDATES
  UPDATES_SIZE = atoi(argv[7]);
#endif
  unsigned First_Time = atoi(argv[argc - 1]);

  if (SAMPLE_PERCENT == 0) 
    {
      SAMPLE_PERCENT = .2;
    }

  if (QUEUE_SIZE == 0) 
    {
      QUEUE_SIZE = 10;
    }

  time_t ltime;
  time(&ltime);
  srand(static_cast<unsigned int>(ltime));

  /*
   * ----------===READ DATASET===----------
   */
  VectString *d_all = new VectString;
  read_dataset(d_all, FILE_DATASET);

  VectString *d;

  if (DATASET_SIZE == 0) 
    {
      d = d_all;
      DATASET_SIZE = d->size();
    }
  else
    {
      d = new VectString(d_all->begin(), d_all->begin() + DATASET_SIZE);
      delete d_all;
    }

  cout << "Dataset size\t" << d->size() << endl;

  if (CLUSTER_NO == 0) 
    {
      CLUSTER_NO = d->size() / 100;
      if (CLUSTER_NO == 0) 
        {
          CLUSTER_NO = 1;
        }
    }

  /*
   * ----------===BUILD/READ CLUSTERS===----------
   */
  Clusters *cs;
  if (READ_CLUSTERS) {
    cs = new Clusters(d);
    cerr << "...read clusters form file \"" << last(FILE_CLUSTERS) << "\"" << endl; 
    ifstream fin(FILE_CLUSTERS.c_str());
    if (!fin) { cerr << "can't open input file \"" << FILE_CLUSTERS << "\"" << endl; exit(EXIT_FAILURE); } 
    fin >> *cs;
    cout << "Clusters read\t" << cs->size_cluster() << endl;
  }
  else {
    cerr << "...build clusters\t" << CLUSTER_NO <<endl;
    clock_t start, ends;
    start = clock();
    cs = new ClustersType(d);
    cs->build_ftables();
    ends = clock();
    cout << "Clusters running time\t"
         << static_cast<double>(ends - start)/CLOCKS_PER_SEC << " "
         << displayTime(static_cast<double>(ends - start)/CLOCKS_PER_SEC) << endl;
    cerr << "...write clusters to file \"" << last(FILE_CLUSTERS) << endl; 
    ofstream fout(FILE_CLUSTERS.c_str());
    if (!fout) { cerr << "can't open output file \"" << FILE_CLUSTERS << "\"" << endl; exit(EXIT_FAILURE); }
    fout << *cs;
    cout << "Clusters write\t" << cs->size_cluster() << endl;
  }

#ifdef CLUST_HIST
#if CLUST_HIST == 1
  {
    //     cerr << "...write clusters hist to file \"" << last(FILE_CLUST_HIST) << "\"..." << endl; 
    //     ofstream fout(FILE_CLUST_HIST.c_str());
    //     if (!fout)
    //       {
    // 	cerr << "can't open output file \"" << FILE_CLUST_HIST << "\"" << endl;
    // 	exit(EXIT_FAILURE);
    //       }
    cout << "Cluster histogram" << endl;
    cs->hist(cout, *d);
    cout << endl;
    //     cout << "Clusters hist write" << endl;
  }
#endif
#endif

#ifdef CLUST_STAT
#if CLUST_STAT == 1
  {
    //     cerr << "...write clusters stat to file \"" << last(FILE_CLUST_STAT) << "\"..." << endl; 
    //     ofstream fout(FILE_CLUST_STAT.c_str(), ios::out | ios::app);
    //     if (!fout)
    //       {
    // 	cerr << "can't open output file \"" << FILE_CLUST_STAT << "\"" << endl;
    // 	exit(EXIT_FAILURE);
    //       }
    cout << "Cluster statistic" << endl;
    cs->stat(cout, *d);
    cout << endl << endl;
    //     cout << "Clusters stat write" << endl;
  }
#endif
#endif

  /*
   * ----------===BUILD/READ PPD-TABLE===----------
   */
  if (READ_PPDTABLE) 
    {
      cerr << "...read ppdtable form file \"" << last(FILE_PPDTABLE) << "\""<< endl;
      ifstream fin(FILE_PPDTABLE.c_str()); // ios::binary
      if (!fin) { cerr << "can't open input file \"" << FILE_PPDTABLE << "\"" << endl; exit(EXIT_FAILURE); } 
      fin >> cs->get_ppdtable(); // .deserialize(fout);
      if (DO_UPDATES) 
        {
          fin >> cs->get_query_pivot();
        }
      cout << "PPDTable read\t" << cs->get_ppdtable().size_table();
      if (DO_UPDATES) 
        {
          cout << " + "<< cs->get_query_pivot().size();
        }
      cout << endl;
    } 
  else 
    {
      cerr << "...build ppdtable" << endl;
    
      unsigned long start, ends;
      start = clock();
      cs->build_ppdtable();
      ends = clock();

      cout << "PPDTable running time\t"
           << static_cast<double>(ends - start)/CLOCKS_PER_SEC << " "
           << displayTime(static_cast<double>(ends - start)/CLOCKS_PER_SEC) << endl;
      cerr << "...write ppdtable to file \"" << last(FILE_PPDTABLE) << "\""<< endl;
      ofstream fout(FILE_PPDTABLE.c_str()); // ios::binary
      if (!fout) { cerr << "can't open output file \"" << FILE_PPDTABLE << "\"" << endl; exit(EXIT_FAILURE); }
      fout << cs->get_ppdtable(); // .serialize(fout);
      fout << endl;
      if (DO_UPDATES) 
        {
          fout << cs->get_query_pivot();
        }
      cout << "PPDTable write\t" << cs->get_ppdtable().size_table();
      if (DO_UPDATES) 
        {
          cout << " + "<< cs->get_query_pivot().size();
        }
      cout << endl;
    }
  
  /*
   * ----------===CREATE/READ RECORDS===----------
   */
  ErrorCorr errorCorr;
  if (DO_ERRORCORR) 
    {
      VectRecord vectRecord;
      unsigned record_size = 0;
      if (READ_RECORDS) {
        read_records(vectRecord,  FILE_RECORDS);
        record_size = static_cast<unsigned>(vectRecord.size());
      }
      else {
        cerr << "...build records" << endl;
        record_size = RECORD_SIZE;
        if (record_size > d->size()) 
          {
            record_size = d->size();
          }

        const unsigned data_size = static_cast<unsigned>(d->size());

        VectUnsigned indexes;
        indexes.reserve(data_size);
        for (unsigned i = 0; i < data_size; i++) indexes[i] = i;

        for (unsigned i = 0; i < record_size; i++) 
          {
            unsigned index = rand() % (data_size - i);
            indexes[index] = indexes[data_size - i - 1];

#if _SimDist_ == 1
            SimType dist = static_cast<SimType>(rand() % (THRESHOLD_MAX - THRESHOLD_MIN + 1) + THRESHOLD_MIN);
#elif _SimDist_ == 2
            SimType dist = static_cast<SimType>(rand() % 
                                                static_cast<int>(THRESHOLD_MAX * 10 - 
                                                                 THRESHOLD_MIN * 10 + 1)) / 10 + THRESHOLD_MIN;
#endif

            Record r = Record((*d)[index], dist);
            r.real = FreqRealFunc(*d, r);

            // 	    if (r.real < REAL_THRESHOLD) {
            // 	      continue;
            // 	    }

            r.est = FreqEstFunc(*d, *cs, r);
            r.setRelErr(); // assumes that the query is form the dataset => real! = 0
            vectRecord.push_back(r);
          }

        cerr << "...write records to file \"" << last(FILE_RECORDS) << "\""<< endl; 
        ofstream fout(FILE_RECORDS.c_str());
        if (!fout) { cerr << "can't open output file \"" << FILE_RECORDS << "\"" << endl; exit(EXIT_FAILURE); }
        fout << vectRecord;
        cout << "Records write\t" << vectRecord.size() << endl;
      }

      /*
       * ----------===BUILD/READ ERROR CORRECTION===----------
       */
      if (READ_ERRORCORR) {
        cerr << "...read errorcorrs form file \"" << last(FILE_ERRORCORR) << "\""<< endl; 

        ifstream fin(FILE_ERRORCORR.c_str());
        if (!fin) { cerr << "can't open input file \"" << FILE_ERRORCORR << "\"" << endl; exit(EXIT_FAILURE); } 

        fin >> errorCorr;
        cout << "ErrorCorr read\t" << vectRecord.size() << endl;
      }
      else {
        errorCorr = ErrorCorr(vectRecord);

        cerr << "...write errorcorr to file \"" << last(FILE_ERRORCORR) << "\""<< endl; 
        ofstream fout(FILE_ERRORCORR.c_str());
        if (!fout) { cerr << "can't open output file \"" << last(FILE_ERRORCORR) << "\"" << endl; exit(EXIT_FAILURE); }
        fout << errorCorr;
        cout << "ErrorCorr write\t" << vectRecord.size() << endl;
      }
    }

  /*
   * ----------===CREATE/READ UPDATES===----------
   */
  if (DO_UPDATES) 
    {
      ContUpdates *contUpdates = new ContUpdates();
      if (READ_UPDATES) 
        {
          cerr << "...read updatess form file \"" << last(FILE_UPDATES) << "\"" << endl; 
          ifstream fin(FILE_UPDATES.c_str());
          if (!fin) { cerr << "can't open input file \"" <<last(FILE_UPDATES) << "\"" << endl; exit(EXIT_FAILURE); } 
          fin >> *contUpdates;
          cout << "Updates read\t" << contUpdates->size() << endl;
          UPDATES_SIZE = contUpdates->size();
        }
      else 
        {
          // CREATE UPDATES
          ContQueryPivot query_pivot = cs->get_query_pivot();
          unsigned pivot = query_pivot.begin()->first;
          set<unsigned> samples;
          for (ContQueryPivot::const_iterator it = query_pivot.find(pivot);
               it != query_pivot.end() && it->first == pivot;
               ++it) 
            {
              samples.insert(it->second.query);
            }

          for (unsigned i = 0; i < UPDATES_SIZE;) 
            {
              unsigned id = rand() % DATASET_SIZE;
              if (cs->is_pivot(id) || samples.find(id) != samples.end()) 
                {
                  continue;
                }
              icstring str = (*d)[id];
              unsigned chg = rand() % MAX_CHANGES;
              EditDist_modif(str, chg);
              contUpdates->push_back(Update(id, str));
              i++;
            }

          cerr << "...write updates to file \"" << last(FILE_UPDATES) << endl; 
          ofstream fout(FILE_UPDATES.c_str());
          if (!fout) { cerr << "can't open output file \"" << last(FILE_UPDATES) << "\"" << endl; exit(EXIT_FAILURE); }
          fout << *contUpdates;
          cout << "Updates write\t" << contUpdates->size() << endl;
        }

      // DO UPDATES
      clock_t start, ends;
      start = clock();
      unsigned i = 0;
      for (ContUpdates::const_iterator it = contUpdates->begin(); it != contUpdates->end(); ++it, i++) 
        {
          // 	  cout << "update\t" << i << "\t" << *it << endl;
          cs->erase(it->id);
          (*d)[it->id] = it->str;
          cs->insert(it->id);
        }
      ends = clock();
      cout << "Updates running time\t"
           << static_cast<double>(ends - start)/CLOCKS_PER_SEC << " "
           << displayTime(static_cast<double>(ends - start)/CLOCKS_PER_SEC) << endl;

      delete contUpdates;
    }

  /*
   * ----------===CREATE/READ TESTS===----------
   */
  bool doBuckets = false;
#ifdef DO_BUCKETS
#if DO_BUCKETS == 1
  doBuckets = true;
#endif
#endif
  Test test = Test(d, cs, doBuckets);
  if (READ_TEST)
    {
      read_test(test, FILE_TEST);
    }
  else
    {
      SetUnsigned setQuery;
      unsigned query_size;
      if (READ_QUERY)
        {
          read_query(setQuery, FILE_QUERY);
          query_size = static_cast<unsigned>(setQuery.size());
        }
      else 
        {
          const unsigned data_size = static_cast<unsigned>(d->size());
          query_size = TEST_SIZE;
          if (query_size > DATASET_SIZE) 
            {
              query_size = DATASET_SIZE;
            }
          while (setQuery.size() < query_size)
            {
              setQuery.insert(rand()%data_size);
            }
          cerr << "...write query to file \"" << last(FILE_QUERY) << endl; 
          ofstream fout(FILE_QUERY.c_str());
          if (!fout) { cerr << "can't open output file \"" << last(FILE_QUERY) << "\"" << endl; exit(EXIT_FAILURE); }
          fout << setQuery;
          cout << "Query write\t" << setQuery.size() << endl;
        }

      test.create(setQuery);
      write_test(test, FILE_TEST);
    }

  /*
   * ----------===TEST===----------
   */
  if (DO_ERRORCORR)
    {
      test.setErrorCorr(errorCorr);
    }
  test.test();

  /*
   * ----------===OUTPUT===----------
   */
  cout << "SEPIA" << endl << "-----" << endl;
  cout << "Dataset size\t" << d->size() << endl;
  cout << endl;
  SimVect::info(cout);
  ClustersType::info(cout);
  PPDSample::info(cout);
  if (DO_UPDATES) 
    {
      cout << "Update size\t" << UPDATES_SIZE << endl << endl;
    }
  if (DO_ERRORCORR)
    {
      ErrorCorr::info(cout);
    }
  test.info(cout);
  cout << "Lookup total\t" << lookup << endl;
  cout << "Lookup hit\t" << lookup_hit << endl;
  cout << "Lookup miss\t" << lookup_miss << endl;  

  cout << endl << endl << endl;

  delete cs;
  delete d;
}
