#include "Sample.h"

#include <fstream>
#include <iomanip>
#include <iostream>

#include "misc.h"

Sample::Sample(): count(0)
{
}

Sample::Sample(unsigned max): count(0)
{
  init_index(max);
  init_rand();
}

Sample::Sample(unsigned no, unsigned max): count(0)
{  
  if (no < max) 
    {
      init_index(max);
      init_rand();

      sample.reserve(no);
      for (; count < no; count++) 
	{
	  unsigned idx = rand() % (max - count);
	  unsigned val = index[idx];
	  index[idx] = index[max - count - 1];
	  sample.push_back(val);
	}
    }
  else
    {
      sample.reserve(no);
      for (; count < no; count++) sample.push_back(count);
    }
}

Sample::Sample(string file_name): count(0)
{
  read(file_name);
}

void Sample::init_index(unsigned max)
{
  index.reserve(max);
  for (unsigned i = 0; i < max; i++) index.push_back(i);
}

void Sample::init_rand() const
{
  time_t ltime;
  time(&ltime);
  srand(static_cast<unsigned int>(ltime));
}

unsigned Sample::generate() 
{
  if (index.size() == count)
    {
      cerr << "no more samples in Sample::generate" << endl;
      exit(1);
    } 
  unsigned idx = rand() % (index.size() - count);
  unsigned val = index[idx];
  index[idx] = index[index.size() - count - 1];
  count++;
  return val;
}

void Sample::push_back(unsigned val)
{
  sample.push_back(val);
}

void Sample::read(string file_name)
{
  cerr << "reading sample form file \"" << last(file_name) << "\"..." << endl; 
  ifstream fin(file_name.c_str());
  if (!fin) {
    cerr << "can't open input file \"" << last(file_name) << "\"" << endl; 
    exit(EXIT_FAILURE); 
  } 
  fin >> *this;
  cerr << "sample read (" << sample.size() << ")" << endl;
}

void Sample::write(string file_name)
{
  cerr << "writing sample to file \"" << last(file_name) << "\"..." << endl; 
  ofstream fout(file_name.c_str());
  if (!fout) { 
    cerr << "can't open output file \"" << last(file_name) << "\"" << endl;
    exit(EXIT_FAILURE);
  }
  fout << *this;
  cerr << "sample (" << sample.size() << ") wrote" << endl;
}
      
ostream& operator<<(ostream& out, const Sample& s)
{
  out << s.sample.size() << endl << endl;
  for (vector<unsigned>::const_iterator it = s.sample.begin(); it != s.sample.end(); ++it)
    out << *it << endl;
  return out;
}

istream& operator>>(istream& in, Sample& s) 
{
  unsigned no;
  in >> no;
  s.sample.reserve(no);
  for (; s.count < no; s.count++) 
    {
      unsigned val;
      in >> val;
      s.sample.push_back(val);
    }
  return in;
}
