#include <vector>
#include <set>
#include <map>
#include <iterator>
#include <string>
#include <iostream>
#include <fstream>
#include <iomanip>

using namespace std;

#include "../setting.h"
#include "../common.h"
#include "SimDist.h"
#include "EditVect.h"
#include "EditVect_ID.h"
#include "EditVect_IS.h"
#include "EditVect_DS.h"
#include "../FreqEst/Predicate.h"
#include "../Util.h"

typedef map<SimType, unsigned, less<SimType> > MapSimTypeUnsigned;
typedef MapSimTypeUnsigned::iterator MapSimTypeUnsignedIt;

ostream& operator<<(ostream& out, MapSimTypeUnsigned& map);

int main() 
{
  string
    file_dataset = "../"+FILE_PATH+FILE_DATASET,
    file_query   = "../"+FILE_PATH+FILE_QUERY,
    file_distrib = "../"+FILE_PATH+FILE_DISTRIB;

  VectString *d=new VectString;
  read_dataset(d, file_dataset);

  SetUnsigned query;
  read_query(query, file_query);

  unsigned count = 0;

  set < unsigned > set_ed;
  set < EditVect > set_ev;
  set < EditVect_ID > set_id;
  set < EditVect_IS > set_is;
  set < EditVect_DS > set_ds;

//   unsigned
//     eq_id = 0, 
//     eq_is = 0, 
//     eq_ds = 0;
  
//   MapSimTypeUnsigned distrib;
//   for (SimType i=SIM_TYPE_MIN; i<=SIM_TYPE_MAX; i=Upper(i)) {
//     distrib[i]=0;
//   }
  
  cout<<"running..."<<endl;
  for (SetUnsignedIt queryIt=query.begin(); queryIt!=query.end(); queryIt++) {
    cout<<"."; cout.flush();
    for (VectStringIt dataIt=d->begin(); dataIt!=d->end(); dataIt++) {
      count++;

      icstring
	a = (*d)[*queryIt],
	b = *dataIt;
      
      set_ed.insert(EditDist(a,  b));
      set_ev.insert(EditVect(a, b));
      set_id.insert(EditVect_ID(a, b));
      set_is.insert(EditVect_IS(a, b));
      set_ds.insert(EditVect_DS(a, b));

//       EditVect vect = EditVect(a, b);
//       if (vect.ins == vect.del) {
// 	eq_id++;
//       }
//       if (vect.ins == vect.sub) {
// 	eq_is++;
//       }
//       if (vect.del == vect.sub) {
// 	eq_ds++;
//       }

//       SimType dist=UpperOrEqual(SimDist(a, b));
//       distrib[dist]++;
    }
  }
  cout<<endl<<"done."<<endl;

  cout << fixed << setprecision(2);
  cout << left << setw(10) << "count:" << right << setw(11) << count << endl;
  cout << left << setw(10) << "set_ed:" << right << setw(11) << set_ed.size() << setw(10) << (set_ed.size() * 10000.0 / count) << "%" << endl;
  cout << left << setw(10) << "set_ev:" << right << setw(11) << set_ev.size() << setw(10) << (set_ev.size() * 10000.0 / count) << "%" << endl;
  cout << left << setw(10) << "set_id:" << right << setw(11) << set_id.size() << setw(10) << (set_id.size() * 10000.0 / count) << "%" << endl;
  cout << left << setw(10) << "set_is:" << right << setw(11) << set_is.size() << setw(10) << (set_is.size() * 10000.0 / count) << "%" << endl;
  cout << left << setw(10) << "set_ds:" << right << setw(11) << set_ds.size() << setw(10) << (set_ds.size() * 10000.0 / count) << "%" << endl;

//   cout << left << setw(10) << "eq_i_d:" << right << setw(11) << eq_id << setw(10) << (eq_id * 100.0 / count) << "%" << endl;
//   cout << left << setw(10) << "eq_i_s:" << right << setw(11) << eq_is << setw(10) << (eq_is * 100.0 / count) << "%" << endl;
//   cout << left << setw(10) << "eq_d_s:" << right << setw(11) << eq_ds << setw(10) << (eq_ds * 100.0 / count) << "%" << endl;

//   cerr<<"writing distributions to file \""<<file_distrib<<"\"..."<< endl; 
//   ofstream fout(file_distrib.c_str());
//   if (!fout) { cerr<<"can't open output file \""<<file_distrib<<"\""<<endl; exit(EXIT_FAILURE); }
//   fout<<fixed<<setprecision(2);
//   fout<<distrib;
//   cerr<<"distributions wrote ("<<static_cast<unsigned int>(distrib.size())<<")"<< endl;

  return 0;
}

ostream& operator<<(ostream& out, MapSimTypeUnsigned& map) 
{
  for (MapSimTypeUnsignedIt it=map.begin(); it!=map.end(); it++) {
    out<<setw(4)<<it->first<<" "<<setw(10)<<it->second<<endl;
  }
  return out;
}

