#include "EditVect.h"
#include "../dataset.h"

EditVect::EditVect(const icstring& a, const icstring& b): ins(0), del(0), sub(0)
{
  // uses O(m) space instead of O(nm) as the classic algorithm
  unsigned n=static_cast<unsigned>(a.length());
  unsigned m=static_cast<unsigned>(b.length());
  if (n==0 || m==0) {
    if (n==0)
      this->ins=m;
    else
      this->del=n;
    return;
  }
  unsigned i, j, d[MAX_STR_LEN][MAX_STR_LEN], mini;
  for (i=0; i<=n; i++) d[i][0]=i;
  for (j=0; j<=m; j++) d[0][j]=j;
  for (i=1; i<=n; i++)
    for (j=1; j<=m; j++)
      d[i][j]=(min(min(
		       d[i-1][j  ]+1,
		       d[i  ][j-1]+1),
		   d[i-1][j-1]+
		   (a[i-1]==b[j-1]?0:1)));
  //for (i=0; i<=n; i++) {
  //	for (j=0; j<=m; j++)
  //		cout<<(int)d[i][j]<<' ';
  //	cout<<'\n';
  //}
  i=n;
  j=m;
  while (!(i==0 && j==0))
    if (i==0 || j==0) // on the edge - there is only one way from here
      if (i==0) {
	this->ins++;
	j--;
      } else {
	this->del++;
	i--;
      }
    else {
      mini=min(min(
		   d[i  ][j-1],
		   d[i-1][j  ]),
	       d[i-1][j-1]);
      if (mini==d[i][j-1] &&
	  d[i][j]==d[i][j-1]+1) { // Insertion
	ins++;
	j--;
      } else if (	mini==d[i-1][j] &&
			d[i][j]==d[i-1][j]+1) {	// Deletion
	del++;
	i--;
      } else { // Substitution
	sub+=d[i][j]-d[i-1][j-1];
	i--;
	j--;
      } 
    }
}

void EditVect::serialize(ofstream &  out) const
{
  out.write(reinterpret_cast<const char *>(&ins), sizeof(unsigned));
  out.write(reinterpret_cast<const char *>(&del), sizeof(unsigned));
  out.write(reinterpret_cast<const char *>(&sub), sizeof(unsigned));
}

void EditVect::deserialize(ifstream &  in) 
{
  in.read(reinterpret_cast<char *>(&ins), sizeof(unsigned));
  in.read(reinterpret_cast<char *>(&del), sizeof(unsigned));
  in.read(reinterpret_cast<char *>(&sub), sizeof(unsigned));
}

ostream& EditVect::info(ostream& out)
{
  out << "Metric" << endl << "---" << endl;
  out << "Edit Distaace" << endl;
  out << "Edit Vector\t< I, D, S >" << endl;
  out << "max string len\t" << MAX_STR_LEN << endl;
  return out << endl;
}
