#ifndef _EditVect_
#define _EditVect_

#include <fstream>
#include <iostream>

using namespace std;

#include "../icstring.h"

class EditVect 
{
  // <I, D, S>
public:
  unsigned ins, del, sub;

  EditVect();
  EditVect(const icstring& a, const icstring& b);

  operator size_t() const;

  unsigned get_dist() const;
  
  void serialize(ofstream &  out) const;
  void deserialize(ifstream &  in);

  static ostream& info(ostream& out);

  friend ostream& operator<<(ostream& out, const EditVect& ev);
  friend istream& operator>>(istream& in, EditVect& ev);
  friend bool operator<(const EditVect& left, const EditVect& right);
  friend bool operator!=(const EditVect& left, const EditVect& right);
};

inline EditVect::EditVect(): ins(0), del(0), sub(0) 
{
}

inline EditVect::operator size_t() const 
{
  // return as 3 digits
  return static_cast<size_t>((ins%10)*100+(del%10)*10+sub%10);
}

inline unsigned EditVect::get_dist() const 
{
  return ins+del+sub;
}

inline ostream& operator<<(ostream& out, const EditVect& ev) 
{
  return out<<ev.ins<<' '<<ev.del<<' '<<ev.sub;
}

inline istream& operator>>(istream& in, EditVect& ev) 
{
  return in>>ev.ins>>ev.del>>ev.sub;
}

inline bool operator<(const EditVect& left, const EditVect& right) 
{
  if (&left==&right)
    return false;
  if (left.ins!=right.ins)
    return left.ins<right.ins;
  if (left.del!=right.del)
    return left.del<right.del;
  return left.sub<right.sub;
}

inline bool operator!=(const EditVect& left, const EditVect& right) 
{
  return left<right || right<left;
}

#endif
