#ifndef _EditVect_DS_
#define _EditVect_DS_

#include "EditVect.h"
#include "../dataset.h"

class EditVect_DS {
  // < D + S, I >
public:
  unsigned del_sub, ins;

  EditVect_DS();
  EditVect_DS(const icstring& a, const icstring& b);

  operator size_t() const;

  unsigned get_dist() const;

  static ostream& info(ostream& out);

  friend ostream& operator<<(ostream& out, const EditVect_DS& ev);
  friend istream& operator>>(istream& in, EditVect_DS& ev);
  friend bool operator<(const EditVect_DS& left, const EditVect_DS& right);
  friend bool operator!=(const EditVect_DS& left, const EditVect_DS& right);
};

inline EditVect_DS::EditVect_DS(): del_sub(0), ins(0) {
}

inline EditVect_DS::EditVect_DS(const icstring& a, const icstring& b) 
{
  EditVect ev=EditVect(a, b);
  del_sub=ev.del+ev.sub;
  ins=ev.ins;
}

inline EditVect_DS::operator size_t() const {
  // return as 2 digits
  return static_cast<size_t>((del_sub%10)*10+ins%10);
}

inline unsigned EditVect_DS::get_dist() const {
  return del_sub+ins;
}

inline ostream& operator<<(ostream& out, const EditVect_DS& ev) {
  return out<<ev.del_sub<<' '<<ev.ins;
}

inline istream& operator>>(istream& in, EditVect_DS& ev) {
  return in>>ev.del_sub>>ev.ins;
}

inline bool operator<(const EditVect_DS& left, const EditVect_DS& right) {
  if (&left==&right)
    return false;
  if (left.del_sub!=right.del_sub)
    return left.del_sub<right.del_sub;
  return left.ins<right.ins;
}

inline bool operator!=(const EditVect_DS& left, const EditVect_DS& right) {
  return left<right || right<left;
}

inline ostream& EditVect_DS::info(ostream& out)
{
  out << "Metric" << endl << "---" << endl;
  out << "Edit Distnace" << endl;
  out << "Edit Vector\t< D + S, I >" << endl;
  out << "max string len\t" << MAX_STR_LEN << endl;
  return out << endl;
}

#endif
