#ifndef _EditVect_ID_
#define _EditVect_ID_

#include "EditVect.h"
#include "../dataset.h"

class EditVect_ID {
  // < I + D, S >
public:
  unsigned ins_del, sub;

  EditVect_ID();
  EditVect_ID(const icstring& a, const icstring& b);

  operator size_t() const;

  unsigned get_dist() const;

  static ostream& info(ostream& out);

  friend ostream& operator<<(ostream& out, const EditVect_ID& ev);
  friend istream& operator>>(istream& in, EditVect_ID& ev);
  friend bool operator<(const EditVect_ID& left, const EditVect_ID& right);
  friend bool operator!=(const EditVect_ID& left, const EditVect_ID& right);
};

inline EditVect_ID::EditVect_ID(): ins_del(0), sub(0) {
}

inline EditVect_ID::EditVect_ID(const icstring& a, const icstring& b) 
{
  EditVect ev=EditVect(a, b);
  ins_del=ev.ins+ev.del;
  sub=ev.sub;
}

inline EditVect_ID::operator size_t() const {
  // return as 2 digits
  return static_cast<size_t>((ins_del%10)*10+sub%10);
}

inline unsigned EditVect_ID::get_dist() const {
  return ins_del+sub;
}

inline ostream& operator<<(ostream& out, const EditVect_ID& ev) {
  return out<<ev.ins_del<<' '<<ev.sub;
}

inline istream& operator>>(istream& in, EditVect_ID& ev) {
  return in>>ev.ins_del>>ev.sub;
}

inline bool operator<(const EditVect_ID& left, const EditVect_ID& right) {
  if (&left==&right)
    return false;
  if (left.ins_del!=right.ins_del)
    return left.ins_del<right.ins_del;
  return left.sub<right.sub;
}

inline bool operator!=(const EditVect_ID& left, const EditVect_ID& right) {
  return left<right || right<left;
}

inline ostream& EditVect_ID::info(ostream& out)
{
  out << "Metric" << endl << "---" << endl;
  out << "Edit Distnace" << endl;
  out << "Edit Vector\t< I + D, S >" << endl;
  out << "max string len\t" << MAX_STR_LEN << endl;
  return out << endl;
}

#endif
