#ifndef _JaccVect_
#define _JaccVect_

#include "JaccDist.h"

class JaccVect {
  // <intersection, union, editdistance>
public:
  unsigned inte, unio, ed;

  JaccVect();
  JaccVect(unsigned i, unsigned u, unsigned e);
  JaccVect(const icstring& a, const icstring& b);

  operator size_t() const;

  float get_dist() const;

  static ostream& info(ostream& out);

  friend ostream& operator<<(ostream& out, const JaccVect& ev);
  friend istream& operator>>(istream& in, JaccVect& ev);
  friend bool operator<(const JaccVect& left, const JaccVect& right);
  friend bool operator!=(const JaccVect& left, const JaccVect& right);
};

inline JaccVect::JaccVect(): inte(0), unio(0), ed(0) {
}

inline JaccVect::JaccVect(unsigned i, unsigned u, unsigned e): inte(i), unio(u), ed(e) {
}

inline JaccVect::operator size_t() const {
  // return as 3 digits
  return static_cast<size_t>((inte%10)*100+(unio%10)*10+ed%10);
}

inline float JaccVect::get_dist() const {
  return 1-static_cast<float>(inte)/unio;
}

inline ostream& operator<<(ostream& out, const JaccVect& jv) {
  return out<<jv.inte<<' '<<jv.unio<<' '<<jv.ed;
}

inline istream& operator>>(istream& in, JaccVect& jv) {
  return in>>jv.inte>>jv.unio>>jv.ed;
}

inline bool operator<(const JaccVect& left, const JaccVect& right) {
  if (left.inte!=right.inte)
    return left.inte<right.inte;
  if (left.unio!=right.unio)
    return left.unio<right.unio;
  return left.ed<right.ed;
}

inline bool operator!=(const JaccVect& left, const JaccVect& right) {
  return left<right || right<left;
}

#endif
