#ifndef _icstring_
#define _icstring_

#include <fstream>
#include <string>
#include <vector> 

using namespace std;

//--- ignore case stirng ---
// replace functions of the standard char_traits<char>
// so that strings behave in a case-insensitive way
struct ignorecase_traits: public char_traits<char> {
  // return whether c1 and c2 are equal
  static bool eq(const char& c1, const char& c2) {
    return toupper(c1)==toupper(c2);
  }
  // return whether c1 is less than c2
  static bool lt(const char& c1, const char& c2) {
    return toupper(c1)<toupper(c2);
  }
  // compare up to n characters of s1 and s2
  static int compare(const char* s1, const char* s2, size_t n) {
    for (size_t i=0; i<n; ++i) {
      if (!eq(s1[i], s2[i])) {
	return lt(s1[i], s2[i])?-1:1;
      }
    }
    return 0;
  }
  //search c in s
  static const char* find(const char* s, size_t n, const char& c) {
    for (size_t i=0; i<n; ++i) {
      if (eq(s[i], c)) {
	return &(s[i]);
      }
    }
    return 0;
  }
};

// define special type such strings
typedef basic_string<char, ignorecase_traits> icstring;
typedef basic_filebuf<char, ignorecase_traits> icfielbuf;
typedef basic_istringstream<char, ignorecase_traits> iicstringstream;
typedef basic_ifstream<char, ignorecase_traits> iicfstream;
/* define an output operator
 * because the traits type is different than that of std::ostram
 */

inline ostream& operator<<(ostream& strm, const icstring& s) {
  // simply convert the icstirng into a normal string
  return strm<<string(s.data(), s.length());
}

inline istream& operator>>(istream& strm, icstring& s) {
  // simply convert the icstirng into a normal string
  string sn;
  strm>>sn;
  s=icstring(sn.data(), sn.length());
  return strm;
}

typedef vector<icstring> VectString;
typedef VectString::iterator VectStringIt;

#endif
